@@ -8,6 +8,7 @@
 """Command line interface to btrsync. Run as main with '-h' for usage information."""
 
 import re
+import os
 import sys
 import fnmatch
 import posixpath
@@ -51,9 +52,9 @@ class CliProgress(sync.ProgressTransfer):
 	def __init__(self, *args, **kwargs):
 		super().__init__(prog_seq='|/-\\', *args, **kwargs)
 
-	async def report_progress(self, cnt, seq):
-		print('\r', next(seq), humanbytes(cnt[0]),
-		      humanbytes((cnt[0] - cnt[1]) / self.period) + '/sec', end='')
+	async def report_progress(self, total, prev, seq):
+		print('\r', next(seq), humanbytes(total),
+		      humanbytes((total - prev) / self.period) + '/sec', end='')
 
 
 class BaseMatch:
@@ -228,11 +229,25 @@ async def dest_root(loc, rootopts={}, rootargs={}):
 	return root, recvpath
 
 
+async def dump_root(odir, inner_coro=None, rootopts={}, rootargs={}):
+	"""Return a ``(root, receive_path)`` pair for dumping a stream to a local file."""
+	if inner_coro is not None:
+		sroot, _ = await inner_coro
+	else:
+		sroot = None
+	return sync.default_root('file-dump')(**rootopts)(odir, subroot=sroot, **rootargs), '.'
+
+
 async def src_root(loc, rootopts={}, rootargs={}):
 	"""Process a source location string, returning a tuple of ``(btrfs_root, matcher_instance)``."""
 	prot, largs, path = parse_root(loc)
+	if prot == 'local' and os.path.isfile(path):
+		prot = 'file'
 	rtype = sync.default_root(prot)(**largs, **rootopts)
-	if await rtype.is_root(path):
+	if prot == 'file':
+		root = rtype(path, **rootargs)
+		matcher = BaseMatch()
+	elif await rtype.is_root(path):
 		if path.endswith('/'):
 			root = rtype(path, **rootargs)
 			matcher = UnderGlob('*')
@@ -245,20 +260,46 @@ async def src_root(loc, rootopts={}, rootargs={}):
 	return root, matcher
 
 
-class CliConfirm:
+def format_transfer(volpaths, parent, destdir, *, verb=False):
+	"""
+	Format the paths that make up a transfer for display on the command line.
+
+	:param volpaths: sequence of paths to send
+	:param parent: parent to use for incremental send, or :const:`None` for full send
+	:param destdir: destination directory used for receive
+	:param verb: if :const:`True` include more details
+	:returns: formatted string
+	"""
+	vpaths = ',\n'.join(volpaths)
+	if verb:
+		return '\n'.join((
+			'',
+			vpaths,
+			'\t' + (f'incremental from {parent}' if parent is not None else 'full'),
+			f'\tinto {destdir}',
+		))
+	else:
+		return vpaths + '\t' + ('full' if parent is None else 'incr') + ' -> ' + destdir
+
+
+class Confirm(sync.ProgressTransfer):
 	"""
 	Handle the UI aspects of confirming a sync via the command line.
 
 	:param src: the SOURCE command line argument currently being processed
 	"""
-	def __init__(self, src):
+	VERBOSE = False
+
+	def __init__(self, src, *args, **kwargs):
+		super().__init__(*args, **kwargs)
 		self._preview = []
 		self.src = src
 
 	async def transf(self, vols, par, src, dst):
 		"""Transfer function as expected by :meth:`btrsync.sync.BtrSync.sync` that only logs transfers."""
-		self._preview.append('; '.join((',\n'.join(v['path'] for v in vols),
-		                           'full transfer' if par is None else f"incremental from {par['path']}")))
+		volpaths, parent, _ = self._sendpaths(vols, par)
+		recvpath = self._recvpath(volpaths)
+		self._preview.append(format_transfer(volpaths, parent, recvpath, verb=self.VERBOSE))
 
 	def head(self):
 		"""Called at the beginning to print a header."""
@@ -274,7 +315,7 @@ class CliConfirm:
 
 	def confirm(self):
 		"""
-		Called after :meth:`.preview` to give final confirmation.
+		Called after :meth:`preview` to give final confirmation.
 
 		:returns: ``'Y'`` to proceed with sync,
 			``'S'`` to skip the current source and continue with the next, and
@@ -289,36 +330,34 @@ class CliConfirm:
 		return r
 
 
-async def do_btrsync(*, srcs, dst, incls, excls, auto, confirm, syncer, syncopts, transfer, srootopts, srootargs, drootopts, drootargs):
+async def do_btrsync(*, src_coros, dst_coro, incls, excls, auto,
+                     confirm, syncer, syncopts, transfer, transopts):
 	"""
 	Perform btrsync from `srcs` to `dst`.
 
-	:param srcs: sequence of source locations
-	:param dst: destination location
+	:param src_coros: sequence of coroutines that return (source_root, matcher) pairs
+	:param dst_coro: coroutine that returns a (destination_root, receive_path) pair
 	:param incls: list of globs matching subvolumes to include in the sync
 	:param excls: list of globs matching subvolumes to exclude from the sync
 	:param auto: if :const:`False` do a dry run, if :const:`None` ask for confirmation, if :const:`True` proceed without asking
-	:param confirm: :class:`.CliConfirm`-like class to handle user interaction for confirmation
+	:param confirm: :class:`.Confirm`-like class to handle user interaction for confirmation
 	:param syncer: :class:`btrsync.sync.BtrSync`-like class to use for sync
 	:param syncopts: keyword arguments to pass to `syncer`
 	:param transfer: :class:`btrsync.sync.Transfer`-like class to use for sync
-	:param srootopts: keyword arguments to pass to source btrfs root class factories
-	:param srootargs: keyword arguments to pass to source btrfs root classes
-	:param drootopts: keyword arguments to pass to the destination btrfs root class factory
-	:param drootargs: keyword arguments to pass to the destination btrfs root class
+	:param transopts: keyword arguments to pass to `transfer`
 	"""
-	dtask = asyncio.create_task(dest_root(dst, drootopts, drootargs))
-	stasks = [asyncio.create_task(src_root(s, srootopts, srootargs)) for s in srcs]
+	dtask = asyncio.create_task(dst_coro)
+	stasks = list(map(asyncio.create_task, src_coros))
 	try:
 		droot, recvpath = await dtask
-		trans = transfer(recvpath=recvpath)
+		trans = transfer(recvpath=recvpath, **transopts)
 		sources = await asyncio.gather(*stasks)
 	except:
 		for t in stasks:
 			t.cancel()
 		await asyncio.wait(stasks)
 		raise
-	for (sroot, matcher), cursrc in zip(sources, srcs):
+	for sroot, matcher in sources:
 		if incls is not None:
 			matcher.include(*incls)
 		matcher.exclude(*excls)
@@ -330,9 +369,9 @@ async def do_btrsync(*, srcs, dst, incls, excls, auto, confirm, syncer, syncopts
 		o.update(syncopts)
 		# Confirmation
 		if auto is not True:
-			conf = confirm(cursrc)
+			conf = confirm(sroot.name, recvpath=recvpath, **transopts)
 			conf.head()
-			if not await s.sync(conf.transf, **o):
+			if not await s.sync(conf, **o):
 				break
 			conf.preview()
 			if auto is False:
@@ -343,47 +382,61 @@ async def do_btrsync(*, srcs, dst, incls, excls, auto, confirm, syncer, syncopts
 			elif cont != 'Y':
 				break
 		# Go time
-		if not await s.sync(trans.transf, **o):
-			break
+		if not await s.sync(trans, **o):
+			raise RuntimeError()
 
 
 def process_args(cliargs):
 	"""Process :mod:`argparse`-style output into arguments for :func:`.do_btrsync`."""
 
-	class CliTransfer(CliProgress if cliargs.progress and not cliargs.quiet else sync.Transfer):
+	prog = cliargs.progress and not cliargs.quiet
+	class CliTransfer(CliProgress if prog else sync.Transfer):
 		"""Transfer class tailored to cli arguments."""
 		if cliargs.quiet < 2:
 			@staticmethod
 			def err(e, *args):
 				print('Error:', e, file=sys.stderr)
-				if args:
-					print(f"@ {', '.join(x['path'] for x in args)}", file=sys.stderr)
 
 		if not cliargs.quiet:
-			if cliargs.verbose:
-				@staticmethod
-				async def report(vols, par, src, dst):
-					print(f"At {', '.join(x['path'] for x in vols)}; ", end='')
-					print('full' if par is None else f"incremental from {par['path']}")
-			else:
-				@staticmethod
-				async def report(vols, par, src, dst):
-					print(f"At {', '.join(posixpath.basename(x['path']) for x in vols)}; " +
-					      ('full' if par is None else 'incremental'))
+			async def report(self, vols, par, src, dst):
+				volpaths, parent, _ = self._sendpaths(vols, par)
+				recvpath = self._recvpath(volpaths)
+				print(format_transfer(volpaths, parent, recvpath, verb=cliargs.verbose))
 			@staticmethod
 			async def report_done(vols, par, src, dst):
 				print(" - Done")
 
-	rootopts = {}
-	if cliargs.sudo:
-		rootopts['sudo'] = cliargs.sudo
-	rootargs = {}
+	class CliConfirm(Confirm):
+		VERBOSE = cliargs.verbose
+
+		if cliargs.quiet < 2:
+			@staticmethod
+			def err(e, *args):
+				print('Error:', e, file=sys.stderr)
+
+	transopts = {'replicate_dirs': cliargs.replicate_dirs}
+	if prog:
+		transopts['period'] = cliargs.progress_period
+
+	srootopts = {'sudo': cliargs.sudo or cliargs.sudo_src}
+	drootopts = {'sudo': cliargs.sudo or cliargs.sudo_dest}
+	srootargs = {}
 	if cliargs.scope is not None:
-		rootargs['scope'] = cliargs.scope
+		srootargs['scope'] = cliargs.scope
+	drootargs = {'create_recvpath': cliargs.create_destpath or cliargs.replicate_dirs}
+	src_coros = [src_root(s, srootopts, srootargs) for s in cliargs.src]
+	dst_coro = dest_root(cliargs.dst, drootopts, drootargs)
+	if cliargs.output_dir or cliargs.output_pipe:
+		dumpargs = drootargs.copy()
+		if cliargs.output_pipe:
+			dumpargs['dump_pipe'] = cliargs.output_pipe
+		if cliargs.output_ext:
+			dumpargs['ext'] = cliargs.output_ext
+		dst_coro = dump_root(cliargs.output_dir, dst_coro, drootopts, dumpargs)
 
 	return {
-		'srcs': cliargs.src,
-		'dst': cliargs.dst,
+		'src_coros': src_coros,
+		'dst_coro': dst_coro,
 		'incls': cliargs.include,
 		'excls': cliargs.exclude,
 		'auto': cliargs.auto,
@@ -391,10 +444,7 @@ def process_args(cliargs):
 		'syncer': IncrSync if cliargs.incremental_only else sync.BtrSync,
 		'syncopts': {'batch': cliargs.batch, 'parallel': cliargs.parallel, 'transfer_existing': cliargs.existing},
 		'transfer': CliTransfer,
-		'srootopts': rootopts,
-		'srootargs': rootargs,
-		'drootopts': rootopts,
-		'drootargs': rootargs
+		'transopts': transopts,
 	}
 
 
@@ -445,8 +495,28 @@ def cli_parser():
 	parser.add_argument('-P', '--parallel', action='store_true',
 	                    help='run independent transfers in parallel')
 
+	parser.add_argument('-o', '--output-dir', metavar='DIR',
+	                    help='dump the send streams into DIR instead of performing a receive at DESTINATION')
+	parser.add_argument('-O', '--output-pipe', metavar='PIPELINE',
+	                    help='''pass the send stream through PIPELINE before dumping to file;
+	                    if supplied without --output-dir the output of PIPELINE is sent to stdout''')
+	parser.add_argument('-e', '--output-ext', metavar='EXT',
+	                    help='(requires --output-dir) append EXT to dump filenames')
+
+	parser.add_argument('-c', '--create-destpath', action='store_true',
+	                    help='create the path specified in DESTINATION if it does not exist')
+	parser.add_argument('-r', '--replicate-dirs', action='store_true',
+	                    help='''(implies `-c') replicate the directory structure
+	                    containing subvolumes in SOURCEs over to DESTINATION;
+	                    paths are taken relative to the source subvolume root
+	                    and applied on top of DESTINATION''')
+
 	parser.add_argument('-s', '--sudo', action='store_true',
-	                    help="use `sudo' for btrfs commands")
+	                    help="use `sudo' for commands, in both source and destination")
+	parser.add_argument('--sudo-src', action='store_true',
+	                    help="use `sudo' for commands executed in source")
+	parser.add_argument('--sudo-dest', action='store_true',
+	                    help="use `sudo' for commands executed in destination")
 	parser.add_argument('--scope', choices=('all', 'strict', 'isolated'),
 	                    help='''set the scope for subvolume discovery:
 	                    'all' considers all accessible subvolumes,
